use crate::malclient::LOCALHOSTV4;
use http_body_util::Full;
use hyper::{body::Bytes, http::Error as HttpError, server::conn::http1, service::service_fn, Request, Response};
use hyper_util::rt::TokioIo;
use log::{debug, error};
use myanimelist::AuthorizationCode;
use queryst::parse;
use tokio::net::TcpListener;

use super::{CALLBACKRESULT, CSRFTOKEN};

#[inline]
fn response(text: &'static str, status: u16) -> Result<Response<Full<Bytes>>, HttpError> {
    Response::builder()
        .status(status)
        .header("Content-Type", "text/plain")
        .body(Full::new(Bytes::from(text)))
}

async fn service(request: Request<hyper::body::Incoming>) -> Result<Response<Full<Bytes>>, HttpError> {
    if request
        .method()
        .as_str()
        != "GET"
        || request
            .uri()
            .query()
            .is_none()
    {
        return response("Sorry this must be redirected from MyAnimelist", 400);
    }
    let token = match CSRFTOKEN.lock() {
        Ok(token) => token,
        Err(error) => {
            error!("Failed to get token: {}", error);
            return response("Internal Server Error", 500);
        },
    };
    let query = match parse(
        request
            .uri()
            .query()
            .unwrap(),
    ) {
        Ok(query) => query,
        Err(error) => {
            error!("failed to parse Query: {}", error.message);
            return response("Sorry this must be redirected from MyAnimelist", 400);
        },
    };
    if let Some(querytoken) = query.get("state") {
        if querytoken
            .as_str()
            .unwrap()
            != token.secret()
        {
            debug!(
                "Querytoken != token: {} != {}",
                querytoken
                    .as_str()
                    .unwrap(),
                token.secret()
            );
            return response("Invalid Token, please retry with the latest displayed link", 400);
        }
    } else {
        return response("Sorry this must be redirected from MyAnimelist", 400);
    }
    let mut callback = CALLBACKRESULT
        .lock()
        .unwrap();
    *callback = AuthorizationCode::new(
        query
            .get("code")
            .unwrap()
            .as_str()
            .unwrap()
            .to_string(),
    );
    response("Successful. You can now close this Tab", 200)
}

pub(super) async fn hyperserver() -> Result<(), ()> {
    let sock = match TcpListener::bind(LOCALHOSTV4).await {
        Ok(sock) => sock,
        Err(error) => {
            error!("Failed to listen ot socket: {}", error);
            return Err(());
        },
    };
    let mut connparser = http1::Builder::new();
    connparser.auto_date_header(true);
    connparser.keep_alive(false);
    loop {
        let stream = match sock
            .accept()
            .await
        {
            Ok((stream, _)) => TokioIo::new(stream),
            Err(error) => {
                error!("Failed to accept connection: {}", error);
                continue;
            },
        };
        match connparser
            .clone()
            .serve_connection(stream, service_fn(service))
            .await
        {
            Ok(_) => {},
            Err(error) => {
                error!("Failed to process request: {}", error);
            },
        };
    }
}
